# Tutorial

```@meta
DocTestSetup = quote
    using PointProcesses
end
```

In this tutorial we demonstrate the main features of PointProcesses.jl.

```jldoctest tuto
julia> using Random

julia> Random.seed!(63);
```

## Working with event histories

To analyze point processes, we need a way to store their realizations. This is the purpose of the [`AbstractHistory`](@ref) subtypes, of which only [`TemporalHistory`](@ref) is implemented for now.

```jldoctest tuto
julia> history = TemporalHistory([0.2, 0.8, 1.1], ["a", "b", "c"], 0.0, 2.0);

julia> duration(history)
2.0

julia> nb_events(history)
3

julia> nb_events(history, 1.0, 2.0)
1

julia> has_events(history)
true

julia> has_events(history, 1.5, 2.0)
false

julia> push!(history, 1.7, "d")

julia> has_events(history, 1.5, 2.0)
true
```

## Working with Markov processes

Some point processes are based on underlying Markov processes, which is why we provide a basic implementation for them.

### Discrete time Markov chains

```jldoctest tuto
julia> dmc = DiscreteMarkovChain(π0 = [0.3, 0.7], P = [0.9 0.1; 0.2 0.8]);

julia> states = rand(dmc, 1000);

julia> dmc_est = fit(DiscreteMarkovChain, states);

julia> round.(dmc_est.P, digits=1)
2×2 Matrix{Float64}:
 0.9  0.1
 0.2  0.8
```

### Continuous time Markov chains

```jldoctest tuto
julia> cmc = ContinuousMarkovChain(π0 = [0.3, 0.7], Q = [-1. 1.; 2. -2.]);

julia> history = rand(cmc, 0., 1000.);

julia> cmc_est = fit(ContinuousMarkovChain, history);

julia> round.(cmc_est.Q, digits=1)
2×2 Matrix{Float64}:
 -1.0   1.0
  1.9  -1.9
```

### Hidden Markov models

## Working with point processes

We finally demonstrate the main goal of the package: point process simulation and inference. All point processes are subtypes of [`AbstractPointProcess{L,M}`](@ref), where `L` is the type of event locations and `M` is the type of event marks.

## Temporal Poisson processes

We provide a number of built-in models, including the basic Poisson process on the real line.

```jldoctest tuto
julia> pp = TemporalPoissonProcess(λ = [0.5, 1., 2.]);

julia> history = rand(pp, 0., 1000.);

julia> pp_init = TemporalPoissonProcess(λ = [1., 1., 1.]);

julia> pp_est = fit(pp_init, history);

julia> round.(pp_est.λ, digits=1)
3-element Vector{Float64}:
 0.5
 1.0
 2.0
```

## Implementing your own temporal point process

To implement your own process, you only have to define a subtype of [`TemporalPointProcess`](@ref) and write the necessary methods: [`intensity`](@ref), [`mark_distribution`](@ref), [`ground_intensity`](@ref) and [`ground_intensity_bound`](@ref).

If these methods exist, the default simulation and inference routines should work, but they can be made much more efficient using custom implementations.